#!/usr/bin/perl
#
# masm2gas - simple-minded MASM-to-GAS converter, for the EPOC hForth project.
# Takes MASM code on standard input, produces GAS code on standard output, with
# errors to standard error.
#
# (C) 2000 Matt J. Gumbley
#
# basic operation:
# ; comment -> @ comment
# 0000h -> 0x0000
# &0000 -> 0x0000
# GBLA -> .global
# MACRO -> .macro
# ENDM -> .endm
# MEND -> .endm
# ORG -> .org
# DB -> .byte
# DW -> .word
# RN -> .req
# SETA -> .set
# x EQU y -> .equ x,y
# EVEN -> .align 4
# $ -> .
#
# also do all the macro processing, since gas 2.6.1 can't cope with this.
use strict;


# Flags that help with debugging
my $remove_comments = 0;
my $debug_macros = 1;
my $debug = 1;

my $inmacro = 0;
my %macros = ();
my $line = 0;
my ($mname, $mparams);

while (<>) {
  $line++;
  chomp;

  # Convert comments
  s-;+([^;]*)-\@$1-;
  s-#.*$-- if ($remove_comments);
  
  # Convert hex numbers
  s/\b([\da-fA-F]+)h(\b|$)/0x$1/g;
  s/&([\da-fA-F]+)/0x$1/g;

  # Convert directives
  s-GBLA-.global-;
  s-DB-.byte-;
  s-DW-.word-;
  s-SETA-.set-;
  s-(\S+)\s+EQU\s+(\S+)(.*)-\t.equ $1, $2 $3-;
  s-MACRO-.macro-;
  s-ENDM-.endm-;
  s-MEND-.endm-;
  s-RN-.req-;
  s-ORG-.org-;
  s-EVEN-.align 4-;

  # $ -> .
  s/\$(\s)/\.$1/g;
  print "[$_]\n" if $debug;

  # Handle macros
  if (/\.macro/) {
    if (/^(\S+)\s+.macro\s*(.*)$/) {
      ($mname, $mparams) = ($1, $2);
      $mparams =~ s/\@.*$//; # remove comments from macro params
      $macros{$mname}{params}=$mparams;
      $macros{$mname}{body}=[];
      $inmacro = 1;
      print "@ MACRO $mname [$mparams] recorded\n" if $debug_macros;
      next;
    }
    else {
      disperr ("WARNING $line: dodgy macro? [$_]\n");
    }
  }
  if (/\.endm/) {
    $inmacro = 0;
    if ($debug_macros) {
      print "@ MACRO $mname body...\n";
      foreach (@{$macros{$mname}{body}}) {
        print "$_\n";
      }
      print "@ MACRO END\n";
    }
    next;
  }
  if ($inmacro) {
    push (@{$macros{$mname}{body}}, $_);
    next;
  }

  # Expand macros repeatedly until they expand no more 
=head
  for (;;) {
    my $latest = expandmacros($_);
    if ($latest eq $_) {
        $_ = $latest;
        last;
    }
    $_ = $latest;
  } 
=cut
  my @processed = expandmacros($_);
  foreach (@processed) {
    print ">>> $_\n";
  }
  print "\n\n" if $debug_macros;
}


sub expandmacros
{
  my $m;
  my $in = shift;
  # NB: $in will always be a single line; it could expand to several
  chomp $in; # shouldn't really be necessary

  # Let's ignore any comment part for now: we'll add it back afterwards
  my $comment;
  ($in, $comment) = stripcomment($in);
  # If there's any WS at end of $in, remove it, but prefix the comment with it
  my ($ws) = $in =~ m/(\s*)$/;
  $ws ||= '';
  $in =~ s/\s*$//;
  $comment = "$ws$comment";
  print "#$in#$comment#\n" if $debug;

  # We produce an array of lines with one macro expanded (there may be more, but
  # we don't deal with them on this pass through).
  my @out = ();
  my $replaced = 0;
  foreach $m (keys %macros) {
    my $qm = quotemeta($m);
    if ($in =~ /$qm/) {
      $replaced = 1;
      if ($debug_macros) {
        print "@ MACRO $m detected...\n";
        print "@ $in\n";
        print "@ MACRO $m prototype [$macros{$m}{params}] body is...\n";
        foreach (@{$macros{$m}{body}}) {
          print "$_\n";
        }
        print "@ MACRO processing begins...\n\n";
      }

      # What parameters have been specified in the macro invocation?
      my ($pa) = $in =~ (/$qm\s+(.*)$/);

      # Macro params can be strings, and in them can be commas, e.g. 'hi,foo'
      # is one parameter.
      # This regex taken from "Mastering Regular Expressions". Thanks Jeffrey!
      my @pas = ();
      push(@pas, $+) while $pa =~ m{
        ('[^\'\\]*(?:\\.[^\'\\]*)*'),?  # groups the phrase inside the quotes
         | ([^,]+),?
         | ,
      }gx;
      push(@pas, undef) if substr($pa,-1,1) eq ',';

      if ($debug_macros) {
        print "@ Input Parameters are:\n";
        foreach (@pas) {
          chomp;
          print "@> [$_]\n";
        }
      }
      # The macro may have been written with params A,B,C, and the call may use
      # a,b,c, so search through the macro text for A, replace it with a; ditto
      # B>b and C>c.
      my @origparams = split(',', $macros{$m}{params}); # no Jeffrey here!
      if ($#pas != $#origparams) {
        disperr ("WARNING $line: macro expansion $m, " . scalar(@pas) .
                 " params; definition had " . scalar(@origparams) . "\n");
        next;
      }
      if ($debug_macros) {
        print "@ MACRO expansion is...\n";
      }
      my $body = $macros{$m}{body}; # an array ref
      # Replace each argument in a copy of the body
      foreach my $line (@{$body}) {
        my $work = $line;
        if ($debug_macros) {
          print "R> [$line]\n";
        }
        foreach my $i (0 .. $#origparams) {
          if ($debug_macros) {
            print "O> [$origparams[$i]] -> [$pas[$i]]\n";
          }
          $work =~ s/$origparams[$i]/$pas[$i]/g; # may need word boundaries?
        }
        if ($debug_macros) {
          print "W> [$work]\n\n";
        }
        push (@out, $work);
      }
      if ($debug_macros) {
        print "@ MACRO expansion end\n";
      }
      last; # remember; only do one macro at a time
    }
  }
  unless ($replaced) {
    $in .= $comment;
    push (@out, $in) if $in;
    return (@out);
  }
  # Append comments...
  push (@out, $comment) if ($comment);
  return (@out);
}

sub disperr
{
  warn @_;
  print @_;
}

sub stripcomment
{
  my $in = shift;
  my $instring = 0;
  my $out = '';
  my $comment = '';
  for my $i (0..length($in)) {
    my $c = substr($in, $i, 1);
    if ($c eq '\'') {
      $instring ^= 1;
    }
    elsif ($c eq '@') {
      unless ($instring) {
        $comment = substr($in, $i, length($in) - $i);
        return ($out, $comment);
      }
    }
    $out .= $c;
  }
  return ($out, '');
}
